This module contains representations and utility functions for batch functions.
The elements of a batch request are described primarily by the type of operation being performed, a user supplied key, and–optionally–a user suppplied value. The value is optional in the case of a delete operation. In industry parlance, the serialized delete operation is a “tombstone”. A sequence number is also associated with the operation. The sequence number is a global, monotonically increasing 64-bit unsigned int. It is never reset.
Structs
- Batch
- A set of operations to perform atomically.
- Batch
Element - Element of a batch operation.