Module config

Module config 

This module contains global configuration constants for RainDB.

These values usually correspond to configurable options for LevelDB (in its options.h file). To get to an MVP and iterate, we keep static values here. These may be made configurable in future versions.

Enums

TableFileCompressionType
The compression types available for blocks within a table file.

Constants

GROUP_COMMIT_SMALL_WRITE_THRESHOLD_BYTES ๐Ÿ”’
The upper threshold for a write to be considered a small write.
ITERATION_READ_BYTES_PERIOD ๐Ÿ”’
The approximate period in bytes between sampling the data read during iteration.
L0_COMPACTION_TRIGGER ๐Ÿ”’
Level-0 compaction is started when we hit this many files.
L0_SLOWDOWN_WRITES_TRIGGER ๐Ÿ”’
Soft limit on the number of level-0 files.
L0_STOP_WRITES_TRIGGER ๐Ÿ”’
Maximum number of level-0 files.
MAX_GROUP_COMMIT_SIZE_BYTES ๐Ÿ”’
The overall maximum group commit batch size.
MAX_MEM_COMPACT_LEVEL ๐Ÿ”’
Maximum level to which a newly compacted memtable is pushed if it does not create an overlap in keys.
MAX_NUM_LEVELS ๐Ÿ”’
The maximum number of SSTable levels that is allowed.
PREFIX_COMPRESSION_RESTART_INTERVAL ๐Ÿ”’
The number of keys between restart points when prefix compressing keys.
SEEK_DATA_SIZE_THRESHOLD_KIB ๐Ÿ”’
The number of kibibytes that are allowed for a single seek of a table file.
SIZE_OF_U32_BYTES ๐Ÿ”’
The size of a u32 in bytes.
SMALL_WRITE_ADDITIONAL_GROUP_COMMIT_SIZE_BYTES ๐Ÿ”’
The allowable additional bytes to add to a group commit where the first writer is doing a small write.
TABLE_FILE_COMPRESSION_TYPE ๐Ÿ”’
The compression type to use for blocks within a table file.