This module contains global configuration constants for RainDB.
These values usually correspond to configurable options for LevelDB (in its options.h file). To
get to an MVP and iterate, we keep static values here. These may be made configurable in future
versions.
Enums
- Table
File Compression Type - The compression types available for blocks within a table file.
Constants
- GROUP_
COMMIT_ ๐SMALL_ WRITE_ THRESHOLD_ BYTES - The upper threshold for a write to be considered a small write.
- ITERATION_
READ_ ๐BYTES_ PERIOD - The approximate period in bytes between sampling the data read during iteration.
- L0_
COMPACTION_ ๐TRIGGER - Level-0 compaction is started when we hit this many files.
- L0_
SLOWDOWN_ ๐WRITES_ TRIGGER - Soft limit on the number of level-0 files.
- L0_
STOP_ ๐WRITES_ TRIGGER - Maximum number of level-0 files.
- MAX_
GROUP_ ๐COMMIT_ SIZE_ BYTES - The overall maximum group commit batch size.
- MAX_
MEM_ ๐COMPACT_ LEVEL - Maximum level to which a newly compacted memtable is pushed if it does not create an overlap in keys.
- MAX_
NUM_ ๐LEVELS - The maximum number of SSTable levels that is allowed.
- PREFIX_
COMPRESSION_ ๐RESTART_ INTERVAL - The number of keys between restart points when prefix compressing keys.
- SEEK_
DATA_ ๐SIZE_ THRESHOLD_ KIB - The number of kibibytes that are allowed for a single seek of a table file.
- SIZE_
OF_ ๐U32_ BYTES - The size of a
u32in bytes. - SMALL_
WRITE_ ๐ADDITIONAL_ GROUP_ COMMIT_ SIZE_ BYTES - The allowable additional bytes to add to a group commit where the first writer is doing a small write.
- TABLE_
FILE_ ๐COMPRESSION_ TYPE - The compression type to use for blocks within a table file.